extern crate getopts;
extern crate gpgme;

use std::env;
use std::fs::File;
use std::io;
use std::io::prelude::*;
use std::process::exit;

use getopts::Options;

use gpgme::{Context, Protocol};

fn print_usage(program: &str, opts: &Options) {
    let brief = format!("Usage: {} [options] FILENAME", program);
    eprintln!("{}", opts.usage(&brief));
}

fn print_result(result: &gpgme::SigningResult) {
    for sig in result.new_signatures() {
        println!("Key fingerprint: {}", sig.fingerprint().unwrap_or("[none]"));
        println!("Signature type : {:?}", sig.mode());
        println!("Public key algo: {}", sig.key_algorithm());
        println!("Hash algo .....: {}", sig.hash_algorithm());
        println!("Creation time .: {:?}", sig.creation_time());
    }
}

fn main() {
    let args: Vec<_> = env::args().collect();
    let program = &args[0];

    let mut opts = Options::new();
    opts.optflag("h", "help", "display this help message");
    opts.optflag("", "openpgp", "use the OpenPGP protocol (default)");
    opts.optflag("", "cms", "use the CMS protocol");
    opts.optflag("", "uiserver", "use the UI server");
    opts.optflag("", "normal", "create a normal signature (default)");
    opts.optflag("", "detach", "create a detached signature");
    opts.optflag("", "clear", "create a clear text signature");
    opts.optopt("", "key", "use key NAME for signing", "NAME");

    let matches = match opts.parse(&args[1..]) {
        Ok(matches) => matches,
        Err(fail) => {
            print_usage(program, &opts);
            eprintln!("{}", fail);
            exit(1);
        }
    };

    if matches.opt_present("h") {
        print_usage(program, &opts);
        return;
    }

    if matches.free.len() != 1 {
        print_usage(program, &opts);
        exit(1);
    }

    let proto = if matches.opt_present("cms") {
        Protocol::Cms
    } else if matches.opt_present("uiserver") {
        Protocol::UiServer
    } else {
        Protocol::OpenPgp
    };

    let mode = if matches.opt_present("detach") {
        gpgme::SignMode::Detached
    } else if matches.opt_present("clear") {
        gpgme::SignMode::Clear
    } else {
        gpgme::SignMode::Normal
    };

    let mut ctx = Context::from_protocol(proto).unwrap();
    ctx.set_armor(true);

    if let Some(key) = matches.opt_str("key") {
        if proto != Protocol::UiServer {
            let key = ctx.find_secret_key(key).unwrap();
            ctx.add_signer(&key).unwrap();
        } else {
            eprintln!("ignoring --key in UI-server mode");
        }
    }

    let mut input = File::open(&matches.free[0]).unwrap();
    let mut output = Vec::new();
    let result = ctx.sign(mode, &mut input, &mut output)
        .expect("signing failed");
    print_result(&result);

    println!("Begin Output:");
    io::stdout().write_all(&output).unwrap();
    println!("End Output.");
}
